from __future__ import with_statement

from mext.classhacks import inject, mix_into, get_name

import sys
from threading import *
from threading import currentThread as current_thread, _active, ThreadError, Event
from thread import get_ident
from mext.context import State
from Queue import Queue
from peak.util.proxies import LazyProxy

from mext.log import Logger

__all__ = ['current_thread', 'is_main_thread', 'Thread', 'SafeLazyProxy',
            'Lock', 'RLock', 'Event', 'Queue']


class SafeLazyProxy(LazyProxy):
    __slots__ = ('__lock__')
    def __init__(self, cb, lock=None):
        SafeLazyProxy.__lock__.__set__(self, lock or RLock())
        LazyProxy.__init__(self, cb)

    @property
    def __subject__(self,
        get_cache=LazyProxy.__cache__.__get__,
        get_subject=LazyProxy.__subject__.fget
    ):
        try:
            return get_cache(self)
        except AttributeError:
            with SafeLazyProxy.__lock__.__get__(self, type(self)):
                return get_subject(self)



class ThreadExt(object):
    mix_into(Thread)
    log = Logger('threads')
    ignore_errors = False
    #if not hasattr(Thread, 'name')
    name = property(Thread.getName, Thread.setName)
    alive = property(lambda self: self.isAlive())
    daemon = property(lambda self: self.isDaemon(), lambda self, d: self.setDaemon(d))
    state = None

    @property
    def is_current(self):
        return current_thread() is self

    @property
    def is_main(self):
        return self.id == main_thread_id

    get_current = staticmethod(current_thread)

    def kill(self):
        self.async_raise(SystemExit)

    @property
    def can_kill(self):
        try:
            from ctypes import pythonapi
            return True
        except ImportError:
            return False

    def async_raise(self, exctype):
        import ctypes
        #thread.get_ident()
        if not issubclass(exctype, BaseException):
            raise TypeError("Only types can be raised (not instances)")
        if not isinstance(self.id, int):
            raise TypeError("tid must be an integer")
        res = ctypes.pythonapi.PyThreadState_SetAsyncExc(self.id, ctypes.py_object(exctype))
        if res == 0:
            raise ValueError("invalid thread id")
        elif res != 1:
            # """if it returns a number greater than one, you're in trouble,
            # and you should call it again with exc=NULL to revert the effect"""
            ctypes.pythonapi.PyThreadState_SetAsyncExc(self.id, 0)
            raise SystemError("PyThreadState_SetAsyncExc failed")

    @classmethod
    def find(self, id):
        return _active.get(id)

    @property
    def id(self):
        if not self.alive:
            raise ThreadError("the thread is not active")
        if not hasattr(self, "_thread_id"):
            if current_thread() is self:
                self._thread_id = get_ident()
            else:
                for tid, tobj in _active.items():
                    if tobj is self:
                        self._thread_id = tid
                        break
                else:
                    raise AssertionError("could not determine the thread's id")
        return self._thread_id

    @classmethod
    def make(cls, func, *args, **kw):
        name = get_name(func, 'Thread-%s' % func)
        return cls(target=func, args=args, kwargs=kw, name=name)

    def start_daemon(self):
        self.daemon = True
        return self.start()




@inject(Thread, save=True)
def start(self): #@@ state arg
    if self.state is None:
        self.state = State.clone()
    #self.log.info("Starting new thread %s", self.name)
    self._Thread_start()
    return self

#@@ Thread.log = Logger('threads[$name/$id]', threads_log)

@inject(Thread, save=True)
def run(self):
##     if self.state is not None:
##         self.state.__#swap()
    self._thread_id = get_ident()
    self.started.set()
    with self.state:
        try:
            return self._Thread_run()
        except SystemExit:
            raise
        except:
            try:
                exc_info = sys.exc_info()
            except AttributeError:
                self.log.warning("Thread raised an exception, but exc_info cannot be aquired")
                return
            if self.ignore_errors:
                #if isinstance(self.ignore_errors, Logger)::
                if self.ignore_errors == 'log':
                    self.log.error("Thread raised an exception", exc_info=True)
                else:
                    self.log.debug("Thread raised an exception %s", exc_info[1])
            else:
                sys.excepthook(*exc_info)
        finally:
            self.stopped.set()
            #self.log.debug("Thread stopped")


@inject(Thread, save=True)
def __init__(self, *args, **kw):
    self._Thread___init__(*args, **kw)
    self.started = Event() #@@xx
    self.stopped = Event()
    #self.state = None#State.get()





assert current_thread().name == 'MainThread'
main_thread_id = get_ident()

def is_main_thread():
    return get_ident() == main_thread_id








